# _*_ coding:UTF-8 _*_
import os
import subprocess
import sys
import time
import traceback
from threading import Thread, Event

from six.moves import queue


class UtilsIO:
    class NonBlockingStreamReader:
        # nbsp = NonBlockingStreamReader(proc.stdout, print_output=True, name="javacap_sever")
        def __init__(self, stream, raise_EOF=False, print_output=True, print_new_line=True, name=None):
            '''
            stream: the stream to read from.
                    Usually a process' stdout or stderr.
            raise_EOF: if True, raise an UnexpectedEndOfStream
                    when stream is EOF before kill
            print_output: if True, print when readline
            '''
            self._s = stream
            self._q = queue.Queue()
            self._lastline = None
            self.name = name or id(self)

            class UnexpectedEndOfStream(Exception):
                pass

            def _populateQueue(stream, queue, kill_event):
                '''
                Collect lines from 'stream' and put them in 'quque'.
                '''
                while not kill_event.is_set():
                    line = stream.readline()
                    if line:
                        queue.put(line)
                        if print_output:
                            # print only new line
                            if print_new_line and line == self._lastline:
                                continue
                            self._lastline = line
                    elif kill_event.is_set():
                        break
                    elif raise_EOF:
                        raise UnexpectedEndOfStream
                    else:
                        # print("EndOfStream: %s" % self.name)
                        break

            self._kill_event = Event()
            self._t = Thread(target=_populateQueue, args=(self._s, self._q, self._kill_event),
                             name="nbsp_%s" % self.name)
            self._t.daemon = True
            self._t.start()  # start collecting lines from the stream

        def readline(self, timeout=None):
            try:
                return self._q.get(block=timeout is not None, timeout=timeout)
            except queue.Empty:
                return None

        def read(self, timeout=0):
            time.sleep(timeout)
            lines = []
            while True:
                line = self.readline()
                if line is None:
                    break
                lines.append(line)
            return b"".join(lines)

        def kill(self):
            self._kill_event.set()


class UtilsOther:
    @staticmethod
    def retries(max_tries, delay=1, backoff=2, exceptions=(Exception,), hook=None):
        def dec(func):
            def f2(*args, **kwargs):
                mydelay = delay
                tries = range(max_tries)
                # support Python conver range obj to list obj
                tries = list(tries)
                tries.reverse()
                for tries_remaining in tries:
                    try:
                        return func(*args, **kwargs)
                    except exceptions as e:
                        if tries_remaining > 0:
                            if hook is not None:
                                hook(tries_remaining, e, mydelay)
                            time.sleep(mydelay)
                            mydelay = mydelay * backoff
                        else:
                            raise

            return f2

        return dec


class UtilsFile:
    @staticmethod
    def read_file(path, encode="utf-8"):
        """
        返回文件内容的字符串，默认使用utf-8编码
        :param path:
        :param encode:
        :return:
        """
        if not os.path.exists(path):
            return ""
        with open(path, "r") as f:
            if encode:
                return f.read(encode)
            else:
                return f.read()


class UtilsProcess:
    @staticmethod
    def execute(command, cwd="/", buf_size=1024):
        """
        执行本地命令
        :param command: 命令
        :param cwd: 执行环境
        :param buf_size: 块大小
        :return: 输出，错误信息(bytes类型)
        """
        sub = subprocess.Popen(command, cwd=cwd, stdin=subprocess.PIPE, stdout=subprocess.PIPE,
                               stderr=subprocess.PIPE, shell=True, bufsize=buf_size)
        (stdout, stderr) = sub.communicate()
        return stdout, stderr

    @staticmethod
    def get_thread_position(thread, model=""):
        frame = sys._current_frames().get(thread.ident, None)
        while True:
            frame = frame
            if frame is None:
                raise RuntimeError("no found target info")
            if model in frame.f_code.co_filename:
                info = traceback.extract_stack(frame)[-1]
                return {"file": info.filename,
                        'line': info.lineno,
                        'code': info._line,
                        'local': frame.f_locals}
            else:
                frame = frame.f_back


class UtilsString:
    @staticmethod
    def get_mid_string(html, start_str, end):
        start = html.find(start_str)
        if start >= 0:
            start += len(start_str)
            end = html.find(end, start)
            if end >= 0:
                return html[start:end].strip()
        return ""

    @staticmethod
    def get_left_string(html, start_str):
        start = html.find(start_str)
        if start >= 0:
            return html[0:start].strip()
        return ""

    @staticmethod
    def get_right_string(html, start_str):
        start = html.find(start_str)
        if start >= 0:
            start += len(start_str)
            return html[start:-1].strip()
        return ""


class UtilsGit:
    def get_git_branch_hex_sha(path):
        def clear_text(s):
            try:
                ret = s.decode('utf-8').strip()
            except:
                ret = ""
            return ret

        stdout, stderr = UtilsProcess.execute("git symbolic-ref --short -q HEAD", path)
        cur_branch = clear_text(stdout)

        stdout, stderr = UtilsProcess.execute("git rev-parse " + cur_branch, path)
        hex_hash = clear_text(stdout)

        stdout, stderr = UtilsProcess.execute("git show " + hex_hash, path)
        info_text = clear_text(stdout)
        lines = info_text.split("\n")

        tag = ""
        for i in range(len(lines) - 1):
            if lines[i] == "":
                tag = lines[i + 1].strip()
                break

        return cur_branch, hex_hash, tag
